arXiv is now an independent nonprofit! Learn more
License: CC BY 4.0
arXiv:2608.12573v2 [cs.LG] 30 Sep 2026

Prof-K: Probabilistic One-Pass Filtering for Efficient Top-kk Selection

Tadeusz Dziarmaga ††thanks: These authors contributed equally to this work. Affiliation: Jagiellonian University    Witold Sikora11footnotemark: 1 Affiliation: Jagiellonian University    Łukasz Struski Affiliation: Jagiellonian University    Jacek Tabor Affiliation: Jagiellonian University    Marcin Mazur Affiliation: Jagiellonian University
Abstract

Top-kk selection is a fundamental computational primitive with applications spanning databases, information retrieval, signal processing, and modern machine learning workloads, including sparse activations and attention pruning. As data sizes grow, existing approaches become inefficient: exact methods incur high memory and compute overhead, while approximate methods often rely on brittle heuristics that degrade under adversarial or heavy-tailed inputs. In this paper, we introduce Prof-K, a fast, scalable, and distribution-agnostic algorithm for exact top-kk selection. Prof-K performs a single-pass filtering procedure: a small random sample estimates an adaptive threshold, the NN input elements are streamed once into a compact buffer, and an exact top-kk routine on this buffer recovers the true top-kk elements on the first attempt with probability at least 1−ε1-\varepsilon, where ε>0\varepsilon>0 is user specified. We derive high-probability guarantees for correctness and buffer size, together with an approximately optimal sample size that minimizes overhead as a function of NN and kk. Empirically, Prof-K achieves 1.5×1.5\times–15×15\times speedups over the highly optimized PyTorch topk and recent RadiK implementations, with the largest gains in the large-scale, small-to-moderate-kk regime where prior methods struggle most. Unlike previous approaches, these guarantees hold independently of the input distribution, ensuring robustness to adversarial settings. A run-time check detects the rare failures and triggers a retry, so the returned set is always exact and ε\varepsilon bounds only the probability of requiring an additional pass. We further demonstrate its impact on training BatchTopK Sparse Autoencoders (SAEs), where top-kk selection constitutes a significant portion of the training cost.

Figure 1: The core idea of our Prof-K algorithm. Top: Full data of size NN sorted in descending order; the true top-kk boundary is marked by a vertical dashed line. Middle: A random sample (solid outlines) is used to compute a threshold τ\tau (green dashed line) as a conservative estimate of the (k/N)(k/N)-quantile. Bottom: The threshold filter is applied in parallel to the full data. Elements ≥τ\geq\tau (green) are retained in a compact candidate buffer, while elements below τ\tau are discarded. This approach ensures all true top-kk elements are preserved with high probability while drastically reducing the workload for the final selection stage.

1 Introduction

Top-kk selection is a core primitive of large-scale machine learning. It arises whenever only a small subset of the largest elements must be retained from a much larger collection [1, 2], for example in sparse autoencoders [3] and gradient sparsification [4]. As tensor sizes continue to grow, the cost of repeated top-kk execution becomes increasingly significant in large-scale GPU workloads, particularly in sparse computation settings where only a small fraction of values is ultimately retained [5, 6, 7, 3].

The challenge is most visible in the large-NN, small-kk regime. Even when only a tiny fraction of elements is needed, exact algorithms must still determine the precise global decision boundary across the full tensor. On modern GPUs, this cost is dominated primarily by memory movement and synchronization, making top-kk disproportionately expensive relative to the surrounding computation [8, 9, 10, 11, 12]. Radix-based kernels such as RadiK [12] reduce the latency but require auxiliary memory proportional to the input, and cannot run at all once the input approaches the device memory. In training workloads, where the selected tensor shares device memory with weights, activations and optimizer state, this overhead directly limits the admissible model or batch size.

In this work, we introduce Prof-K, a probabilistic one-pass filtering algorithm for exact top-kk selection on GPUs. The method draws a small uniform sample to estimate a conservative threshold, streams once through the input while retaining only the elements above this threshold in a compact candidate buffer, and applies an exact top-kk routine to the candidates (Figure 1). The filter kernel counts the retained elements exactly, so the rare case in which the buffer is inadequate is detected at run time and handled by redrawing the sample: the output is always the exact top-kk, and the user-specified budget ε\varepsilon bounds the frequency of the additional pass rather than the accuracy of the result. This distinguishes Prof-K from approximate top-kk methods, which trade accuracy for speed [13]. Because the analysis depends only on the ranks induced by uniform sampling, the guarantee holds for every input, including heavy-tailed and adversarial ones.

We summarize our contributions as follows.

  • •

    We propose Prof-K, a sampling-based filtering algorithm for exact top-kk selection over the large tensors that arise in sparse ML workloads. It always returns the exact result: failures are detected at run time and resolved by resampling, with the budget ε\varepsilon bounding only their frequency, and the guarantee holds for every input distribution. We derive the optimal parameters for a prescribed failure budget, together with closed-form approximations that are accurate and cheap to evaluate.

  • •

    We evaluate Prof-K, implemented in Triton, against torch.topk and RadiK, the state-of-the-art radix selection, on a DGX H100 and an RTX 3060. Prof-K is the fastest method in every large-NN configuration on both GPUs, by up to 16×16\times over torch.topk and 3.1×3.1\times over RadiK, and its speed is unaffected by the input distribution, including inputs adversarial for radix selection. Its auxiliary memory is measured in megabytes where RadiK requires gigabytes, which extends exact selection to inputs on which radix selection cannot run, allowing to train larger models with larger batches on the same hardware.

  • •

    We demonstrate the utility of Prof-K in two machine learning workloads. In BatchTopK sparse autoencoders [3] trained on GPT-2 XL activations it reduces the end-to-end step time by 1.201.20 to 1.30×1.30\times, close to the ceiling attainable by any selection kernel; in top-kk gradient sparsification of GPT-2 XL it reduces the step time by 1.38×1.38\times with results identical to torch.topk. The same batch-global selection structure appears in Matryoshka and crosscoder variants of BatchTopK [14, 15].

2 Related Work

Classical exact top-kk methods are based on sorting, heaps, or partition-based selection such as quickselect, and efficient top-kk query execution has long been studied in database and parallel hardware settings [1, 2]. In large-scale GPU workloads the dominant cost is memory movement rather than comparisons, which makes exact top-kk expensive even in optimized libraries such as PyTorch.

A substantial line of work studies efficient exact top-kk selection on GPUs. Early work by Alabi et al. [8] developed GPU-specific kk-selection algorithms optimized for massively parallel architectures. Dr. Top-k [9] introduces a delegate-centric design that improves load balancing and reduces redundant work among GPU threads. Zhang et al. [10] provide a comprehensive study of parallel GPU top-kk algorithms and propose optimized implementations for different operating regimes. RTop-K [11] focuses on efficient row-wise top-kk selection and demonstrates strong performance for moderate-size neural network workloads. Radix-based approaches such as RadiK [12] further exploit floating-point representations to accelerate exact selection. In contrast, Prof-K reduces the size of the exact selection problem itself through probabilistic filtering, making it particularly effective for very large NN and relatively small kk.

Approximate and probabilistic selection methods reduce computational cost using sampling, sketching, or approximate quantile estimation, often at the cost of weaker guarantees or assumptions on the input distribution. Quantile summary methods such as the Greenwald–Khanna [16] algorithm provide efficient streaming approximations with bounded error guarantees. Prof-K differs in that its guarantees depend only on the combinatorial properties of uniform sampling without replacement. The analysis therefore depends on ranks rather than values, yielding distribution-agnostic probabilistic guarantees.

Top-kk selection is increasingly important in sparse machine learning workloads, including mixture-of-experts routing, sparse attention, activation pruning, and sparse autoencoders [5, 6, 7, 3, 14, 15]. Prof-K targets the subset of these workloads that perform batch-global selection over very large tensors with small retained fractions. Rather than replacing exact selection entirely, Prof-K uses sampling to construct a conservative threshold and then applies exact top-kk refinement on a compact candidate buffer, making it complementary to optimized GPU kernels such as Dr. Top-k, RTop-K, and RadiK.

3 Problem Setup and Algorithm Overview

Given an input vector x∈ℝNx\in\mathbb{R}^{N} and an integer k≥1k\geq 1, our goal is to return the kk largest entries of xx together with their indices. For batched inputs with batch size BB, the procedure is applied in parallel to each batch element.

The Prof-K Algorithm

Prof-K proceeds in four stages, designed to minimize memory traffic while providing probabilistic correctness guarantees:

  1. 1.

    Sampling. Uniformly sample SS indices without replacement from [N]={1,…,N}[N]=\{1,\dots,N\}. Let 𝒮⊂[N]\mathcal{S}\subset[N] denote the sampled indices.

  2. 2.

    Threshold Estimation. Sort the sampled values x|𝒮x|_{\mathcal{S}} in descending order. For a chosen rank t∈[S]t\in[S], define the threshold τ=st\tau=s_{t}, where sts_{t} is the tt-th largest sampled value.

  3. 3.

    Filtering. Stream through the full input once, appending every element satisfying xi≥τx_{i}\geq\tau to a candidate buffer ℬ\mathcal{B} of capacity MM, which is usually small relative to the full input size NN.

  4. 4.

    Refinement. Run an exact top-kk routine on the buffer ℬ\mathcal{B} to recover the final answer. If fewer than kk candidates are collected, or if the buffer overflows before the scan completes, Prof-K redraws the sample and repeats the pass. Each attempt fails with probability at most ε\varepsilon independently of the previous ones, so the expected number of attempts is at most 1/(1−ε)1/(1-\varepsilon); an implementation may also fall back to an exact top-kk over the full input after a fixed number of failed attempts, say six, an event of probability at most ε6\varepsilon^{6}.

The central design goal is to choose the parameters (S,t,M)(S,t,M) so that the number of retained elements R=|{i∈[N]:xi≥τ}|R=|\{i\in[N]:x_{i}\geq\tau\}| satisfies k≤R≤Mk\leq R\leq M with probability at least 1−ε1-\varepsilon, where 0<ε≪10<\varepsilon\ll 1 is user specified. When this holds, the final exact top-kk operates on a small buffer rather than all NN entries. The idea behind the threshold is that the (t/S)(t/S)-quantile of a uniform sample estimates the (t/S)(t/S)-quantile of the input, so choosing tt slightly above S​k/NSk/N yields a conservatively low threshold that all true top-kk elements pass with high probability while most other elements are excluded.

4 Theoretical Analysis

We now formalize the probabilistic guarantees underlying Prof-K. The analysis proceeds in four steps. First, we characterize the distribution of the rank that the sampled threshold occupies in the full input. Second, we prove a theorem that yields the exact optimal threshold rank t⋆t^{\star} and buffer capacity M⋆M^{\star} as functions of NN, kk, SS and ε\varepsilon, without any approximation. Third, we give accurate and inexpensive approximations to these two quantities for settings in which the parameters must be recomputed frequently. Fourth, we derive an asymptotically optimal choice of the sample size SS as a function of NN and kk that minimizes the computational overhead of the method. All proofs are given in Appendix A.

Notation and Conventions

Throughout this section the input x∈ℝNx\in\mathbb{R}^{N} is fixed and its entries are assumed to be distinct, so that every entry has a well-defined rank, with rank 11 assigned to the largest entry. For K∈{0,1,…,N}K\in\{0,1,\dots,N\} we write TK⊂[N]T_{K}\subset[N] for the set of indices of the KK largest entries of xx. The sampled index set 𝒮⊂[N]\mathcal{S}\subset[N] is drawn uniformly at random among all subsets of [N][N] of size SS; this is what sampling SS indices uniformly without replacement means. As in Section 3, τ\tau denotes the tt-th largest sampled value and RR its rank in the full input, so that R=|{i∈[N]:xi≥τ}|R=|\{i\in[N]:x_{i}\geq\tau\}| is the number of elements that pass the filter.

For integers 0≤K≤N0\leq K\leq N and 0≤S≤N0\leq S\leq N, X∼Hypergeometric⁡(N,K,S)X\sim\Hypergeometric(N,K,S) is the number of distinguished items among SS items drawn uniformly without replacement from NN items of which KK are distinguished, that is,

Pr⁡(X=j)=(Kj)​(N−KS−j)(NS),j∈{0,1,…,S},\Pr(X=j)=\frac{\binom{K}{j}\binom{N-K}{S-j}}{\binom{N}{S}},\qquad j\in\{0,1,\dots,S\}, (1)

with (ab)=0\binom{a}{b}=0 whenever b<0b<0 or b>ab>a. We write FN,K,S​(j)=Pr⁡(X≤j)F_{N,K,S}(j)=\Pr(X\leq j) for its distribution function and

FN,K,S−1​(p)=min⁡{j∈{0,1,…,S}:FN,K,S​(j)≥p},p∈(0,1],F^{-1}_{N,K,S}(p)=\min\{\,j\in\{0,1,\dots,S\}:F_{N,K,S}(j)\geq p\,\},\qquad p\in(0,1], (2)

for its quantile function, which takes a probability and returns an integer; both are available in standard numerical libraries.

Sampled Counts in a Top-KK Set

The single quantity on which the analysis rests is the number of sampled indices that fall into a top-KK set. For K∈{0,1,…,N}K\in\{0,1,\dots,N\} let XK=|𝒮∩TK|X_{K}=|\mathcal{S}\cap T_{K}|, and define

G⁡(K)=Pr⁡(fewer than t of the S sampled indices lie in TK)=Pr⁡(XK≤t−1).G(K)=\Pr\bigl(\text{fewer than $t$ of the $S$ sampled indices lie in $T_{K}$}\bigr)=\Pr(X_{K}\leq t-1). (3)

The function GG depends on NN, SS and tt as well, and we make this dependence explicit by writing G⁡(N,K,S,t)G(N,K,S,t) whenever it matters.

Lemma 1.

For every K∈{0,1,…,N}K\in\{0,1,\dots,N\}, XK∼Hypergeometric⁡(N,K,S)X_{K}\sim\Hypergeometric(N,K,S) and consequently

G⁡(K)=FN,K,S​(t−1).G(K)=F_{N,K,S}(t-1). (4)

Moreover, G⁡(K)G(K) is nonincreasing in KK and nondecreasing in tt.

Failure Modes

Exactly RR elements pass the filter. If R≥kR\geq k, the kk-th largest entry of xx is at least τ\tau, so every true top-kk entry passes; if in addition R≤MR\leq M, every element that passes is stored, and the refinement step returns the exact top-kk. Prof-K verifies the condition k≤R≤Mk\leq R\leq M at run time, since the filter kernel counts the elements above the threshold exactly, and treats its violation as a failure. There are two ways in which this can happen:

  1. 1.

    Insufficient recall, A={R<k}A=\{R<k\}: the threshold is too high and fewer than kk elements pass the filter.

  2. 2.

    Buffer overflow, B={R>M}B=\{R>M\}: more elements pass the filter than the buffer can hold, so some true top-kk entries may fail to enter it.

Since M≥kM\geq k, the two events are disjoint and Pr⁡(A∪B)=Pr⁡(A)+Pr⁡(B)\Pr(A\cup B)=\Pr(A)+\Pr(B). An overflow does not by itself make the output wrong, since the true top-kk entries may still be among the MM stored elements, so the probability of an incorrect output is bounded by Pr⁡(A)+Pr⁡(B)\Pr(A)+\Pr(B). We now compute both terms exactly.

Proposition 2 (Insufficient recall).

Let 1≤k≤N1\leq k\leq N and 1≤t≤S1\leq t\leq S. Then for every input xx,

Pr⁡(R<k)=1−G⁡(k−1)=1−FN,k−1,S​(t−1).\Pr(R<k)=1-G(k-1)=1-F_{N,k-1,S}(t-1). (5)
Proposition 3 (Buffer overflow).

Let 1≤t≤S1\leq t\leq S and k≤M≤Nk\leq M\leq N. Then for every input xx,

Pr⁡(R>M)=G⁡(M)=FN,M,S​(t−1).\Pr(R>M)=G(M)=F_{N,M,S}(t-1). (6)

Neither probability depends on the values of xx: only NN, kk, SS, tt and MM enter Equations 5 and 6. This is the precise sense in which the guarantees of Prof-K are distribution-free.

Exact Optimal Parameters

Given a failure budget for each mode, Propositions 2 and 3 reduce the choice of tt and MM to two inequalities in GG, and the monotonicity established in Lemma 1 allows both to be solved exactly. The following theorem gives the smallest threshold rank and buffer capacity under which Prof-K retrieves the exact top-kk with a user-specified probability on the first attempt.

Theorem 4 (Exact optimal parameters).

Let 1≤k≤N1\leq k\leq N, 1≤S≤N1\leq S\leq N and εA,εB∈(0,1)\varepsilon_{A},\varepsilon_{B}\in(0,1) with εA≥Pr⁡(Xk−1=S)\varepsilon_{A}\geq\Pr(X_{k-1}=S), and set

t⋆\displaystyle t^{\star} =FN,k−1,S−1​(1−εA)+1,\displaystyle=F^{-1}_{N,k-1,S}(1-\varepsilon_{A})+1, (7)
M⋆\displaystyle M^{\star} =min⁡{M∈{k,…,N}:FN,M,S​(t⋆−1)≤εB}.\displaystyle=\min\bigl\{\,M\in\{k,\dots,N\}:F_{N,M,S}(t^{\star}-1)\leq\varepsilon_{B}\,\bigr\}.

Then 1≤t⋆≤S1\leq t^{\star}\leq S, the minimum defining M⋆M^{\star} exists, and for every input xx:

  1. (i)

    t⋆t^{\star} is the smallest threshold rank t∈[S]t\in[S] for which the recall failure has probability Pr⁡(R<k)≤εA\Pr(R<k)\leq\varepsilon_{A};

  2. (ii)

    with t=t⋆t=t^{\star}, M⋆M^{\star} is the smallest buffer capacity M≥kM\geq k for which the overflow has probability Pr⁡(R>M)≤εB\Pr(R>M)\leq\varepsilon_{B};

  3. (iii)

    with t=t⋆t=t^{\star} and M=M⋆M=M^{\star}, Prof-K passes its run-time check, and hence returns the exact top-kk, with probability at least 1−εA−εB1-\varepsilon_{A}-\varepsilon_{B}; the exact failure probability is

    Pr⁡(A∪B)=Pr⁡(R<k)+Pr⁡(R>M⋆).\Pr(A\cup B)=\Pr(R<k)+\Pr(R>M^{\star}). (8)

The condition on εA\varepsilon_{A} only excludes a degenerate case: Xk−1=SX_{k-1}=S is the event that the entire sample falls within the top k−1k-1 entries, and if it had probability above εA\varepsilon_{A}, no threshold rank t≤St\leq S could meet the recall budget. Its probability (k−1S)/(NS)\binom{k-1}{S}/\binom{N}{S} is zero for S≥kS\geq k and at most ((k−1)/N)S((k-1)/N)^{S} otherwise, far below any budget of interest. With the default split εA=εB=ε/2\varepsilon_{A}=\varepsilon_{B}=\varepsilon/2, Theorem 4 states that Prof-K run with the parameters (S,t⋆,M⋆)(S,t^{\star},M^{\star}) returns the exact top-kk on the first attempt with probability at least 1−ε1-\varepsilon, for every input. The two parameters are also jointly optimal in the following sense.

Corollary 5 (Joint optimality).

Under the assumptions of Theorem 4, every pair (t,M)(t,M) with t∈[S]t\in[S] and M∈{k,…,N}M\in\{k,\dots,N\} satisfying Pr⁡(R<k)≤εA\Pr(R<k)\leq\varepsilon_{A} and Pr⁡(R>M)≤εB\Pr(R>M)\leq\varepsilon_{B} has t≥t⋆t\geq t^{\star} and M≥M⋆M\geq M^{\star}.

Computing the pair is inexpensive: t⋆t^{\star} is one quantile evaluation, and since GG is nonincreasing in MM, M⋆M^{\star} is found by binary search over {k,…,N}\{k,\dots,N\} with at most ⌈log2⁡N⌉\lceil\log_{2}N\rceil evaluations of the distribution function.

Fast Approximate Parameter Computation

Evaluating Equation 7 costs about a millisecond with a numerical library, comparable to the kernel itself. In most machine learning workloads the input shape and the sparsity level are fixed throughout training, so (t⋆,M⋆)(t^{\star},M^{\star}) are computed once per configuration and cached; our implementation memoizes them. For configurations that change from call to call, Appendix B derives closed-form surrogates (a skewness-corrected normal or a Poisson quantile for t⋆t^{\star}, a Wilson–Hilferty quantile for M⋆M^{\star}) that need only elementary arithmetic, are conservative in every configuration tested, and lie within a few percent of M⋆M^{\star}; Algorithm 1 states the complete procedure.

Algorithm 1 Fast approximate parameter computation (budget split εA=εB=ε/2\varepsilon_{A}=\varepsilon_{B}=\varepsilon/2)
0:  NN, kk, failure budget ε\varepsilon; sample size bounds SminS_{\min}, SmaxS_{\max}; calibrated cost ratio r⁡(⋅)r(\cdot)
0:  sample size SS, threshold rank tt, buffer capacity MM
1:  z←Φ−1​(1−ε/2)z\leftarrow\Phi^{-1}(1-\varepsilon/2) {rational approximation}
2:  κ←z2+12\kappa\leftarrow z^{2}+\tfrac{1}{2}; r←r⁡(k)r\leftarrow r(k) {calibrated cost ratio}
3:  S1←(z/r)2/3​(k​N)1/3S_{1}\leftarrow(z/r)^{2/3}(kN)^{1/3}; S2←(κ​N/r)1/2S_{2}\leftarrow(\kappa N/r)^{1/2} {Proposition 7}
4:  S←clip⁡(⌊max⁡(S1,S2)⌋,Smin,Smax)S\leftarrow\mathrm{clip}\bigl(\lfloor\max(S_{1},S_{2})\rfloor,\;S_{\min},\;S_{\max}\bigr) {SminS_{\min} if r=∞r=\infty or k/N>2−10k/N>2^{-10}}
5:  μ←(k−1)​S/N\mu\leftarrow(k-1)S/N
6:  if μ≥15\mu\geq 15 then
7:   σ←μ⁡(1−k−1N)​N−SN−1\sigma\leftarrow\sqrt{\mu\,(1-\tfrac{k-1}{N})\,\tfrac{N-S}{N-1}}; t←⌈μ+z​σ+(z2−1)/6⌉+1t\leftarrow\lceil\mu+z\sigma+(z^{2}-1)/6\rceil+1 {Equation 10}
8:  else
9:   p←e−μp\leftarrow e^{-\mu}; c←pc\leftarrow p; j←0j\leftarrow 0
10:   while c<1−ε/2c<1-\varepsilon/2 do
11:    j←j+1j\leftarrow j+1; p←p​μ/jp\leftarrow p\,\mu/j; c←c+pc\leftarrow c+p {Poisson quantile, Equation 11}
12:   end while
13:   t←j+1t\leftarrow j+1
14:  end if
15:  h←1/(9​t)h\leftarrow 1/(9t); M←⌈N+1S+1​t​(1−h+z​h)3⌉M\leftarrow\bigl\lceil\tfrac{N+1}{S+1}\,t\,(1-h+z\sqrt{h})^{3}\bigr\rceil {Wilson–Hilferty, Equation 12}
16:  return S,t,MS,\,t,\,M

Choice of the Sample Size

Theorem 4 fixes t⋆t^{\star} and M⋆M^{\star} for a given SS but leaves SS free, and SS controls a trade-off: a larger sample makes the threshold more accurate and the buffer smaller, at the price of more sampling work. We choose SS by minimizing a cost model of the stages of Prof-K that charges the random gather of the SS sampled elements and the selection of the top tt among them against the exact selection over the M⋆​(S)M^{\star}(S) buffer slots; the filtering pass streams the input once whatever SS is. With the leading-order behavior M⋆​(S)≈k+2​z​k​N/S+κ​N/SM^{\star}(S)\approx k+2z\sqrt{kN/S}+\kappa N/S, where z=Φ−1​(1−ε/2)z=\Phi^{-1}(1-\varepsilon/2) for the standard normal distribution function Φ\Phi and κ=z2+12\kappa=z^{2}+\frac{1}{2}, the model has a closed-form minimizer up to a factor below 1.61.6 (Proposition 7 in Appendix B): the optimal sample size scales as (k​N)1/3(kN)^{1/3} for large kk and as N\sqrt{N} below, and its single hardware constant, the ratio rr between the per-element costs of sampling and of selecting from the buffer, is calibrated once per GPU (Appendix E.1). The resulting rule, clipped to [Smin,Smax][S_{\min},S_{\max}] to reflect the fixed launch cost of the selection and to keep the sampling stage negligible against the scan, is given in lines 3 and 4 of Algorithm 1, which collects the complete parameter computation. Since the latency is flat over a wide range of SS around the optimum (Appendix E.1), sweeping SS and ε\varepsilon once for a fixed setup and caching the best values is an equally legitimate choice when the configuration does not change.

5 Experiments and Evaluation

We compare Prof-K against torch.topk and RadiK [12], the state-of-the-art radix selection and the strongest exact baseline at large NN, on a data-centre GPU (NVIDIA DGX H100, 80 GB) and a consumer GPU (NVIDIA RTX 3060 Laptop, 6 GB). Prof-K runs at ε=10−3\varepsilon=10^{-3} with parameters computed once per configuration and cached. The grid covers NN from 2212^{21} to 2302^{30} and kk from 252^{5} to 2192^{19} on the H100, extended to N=231N=2^{31} against torch.topk alone, and NN up to 2282^{28} and kk up to 2182^{18} on the RTX 3060, whose memory limits the input size. Inputs are standard normal float32 tensors; latencies are means over 3030 input tensors and 3030 timed calls each, measured with CUDA events after a warm-up, and auxiliary memory is the peak allocation above the baseline during one call. Other input distributions, including ones adversarial for radix selection, are covered in Appendix E.3, further details in Appendix D, and ablations on batched inputs in Appendix E.

Refer to caption
Refer to caption
Refer to caption
Refer to caption
Figure 2: Speed-up of Prof-K over torch.topk (left) and RadiK (right) across the (N,k)(N,k) grid at ε=10−3\varepsilon=10^{-3} on the DGX H100 (top) and the RTX 3060 (bottom). Blue marks configurations where Prof-K is faster, red where it is slower.

Speed-up Across the (N,k)(N,k) Space

Figure 2 shows the speed-up over the full grid and Figure 3 the absolute latencies at k=213k=2^{13}. Three regimes are apparent. Below N≈223N\approx 2^{23} every method takes a fraction of a millisecond, on the H100 between 0.20.2 and 0.50.5 ms at k=213k=2^{13}, and fixed costs dominate: Prof-K is on par with torch.topk and behind RadiK for small kk on the H100, and ahead of both on the RTX 3060. From N≈224N\approx 2^{24} the streaming pass dominates every method, and the advantage of a single bandwidth-bound pass becomes visible. On the H100 at k=213k=2^{13} the latency of Prof-K grows from 0.250.25 ms at N=224N=2^{24} to 1.71.7 ms at N=230N=2^{30}, against 0.450.45 to 26.626.6 ms for torch.topk and 0.420.42 to 5.45.4 ms for RadiK: the speed-up over torch.topk rises from 1.8×1.8\times at 2242^{24} to 5.2×5.2\times at 2262^{26}, 11×11\times at 2282^{28} and 15×15\times at 2302^{30}, and Prof-K is ahead of RadiK from N=225N=2^{25} on in every configuration but two, reaching 3.1×3.1\times at 2302^{30}. On the RTX 3060 the picture is the same at a smaller scale: at N=228N=2^{28} Prof-K takes 3.93.9 ms against 22.122.1 ms for torch.topk and 10.710.7 ms for RadiK, 5.7×5.7\times and 2.7×2.7\times. Finally, the dependence on kk is weak: at fixed NN the speed-up stays within a factor of about 1.31.3 of its small-kk value up to k≈2−10​Nk\approx 2^{-10}N and declines only beyond that, where the buffer and the refinement grow with kk.

Figure 3: Latency as a function of the input size at k=213k=2^{13} on both GPUs (log–log). Both baselines perform several passes over the input, Prof-K a single one, so its slope is that of the memory bandwidth. On the H100 the grid extends to N=231N=2^{31} beyond the 3232-bit limit of RadiK.

Memory Footprint

Figure 4 reports the auxiliary memory of the three methods on the H100 at k=213k=2^{13}. Both baselines are O⁡(N)O(N): RadiK allocates a workspace of about 88 bytes per input element, twice the size of the input itself, from 1616 MB at N=221N=2^{21} to 88 GB at N=230N=2^{30}, and torch.topk about 3232 MB at 2302^{30}. The workspace of RadiK is a structural property of its radix selection, which must keep per-element state across its passes. Prof-K allocates only the sample, the candidate buffer and the output, so its footprint is O⁡(M)O(M) rather than O⁡(N)O(N), and the measurements confirm it: from 0.80.8 MB at N=221N=2^{21} to 44 MB at N=230N=2^{30} at k=213k=2^{13}, growing with M⋆M^{\star} and not with the input, and between 22 and 55 MB at N=230N=2^{30} for every k≤213k\leq 2^{13}, exceeding torch.topk only when kk approaches 2202^{20}. Prof-K therefore never runs out of memory where the input itself fits, and it keeps scaling where radix selection stops: at N=231N=2^{31} it selects the top 2132^{13} elements in 3.13.1 ms against 53.353.3 ms for torch.topk, a 17×17\times speed-up, and it stays above 13×13\times up to k=219k=2^{19} and above 6×6\times at k=222k=2^{22} for inputs of 3⋅1093\cdot 10^{9} and 4.3⋅1094.3\cdot 10^{9} elements (Table 4 in Appendix D).

Figure 4: Auxiliary GPU memory on the DGX H100 at k=213k=2^{13} (log scale); the dotted line is the size of the input. The footprints of torch.topk and RadiK grow with NN, that of Prof-K with the buffer capacity M⋆M^{\star}.

Application: BatchTopK Sparse Autoencoders

To test whether kernel-level gains survive inside a training loop, we replace the batch-global top-kk of BatchTopK sparse autoencoders (SAEs) [3] with Prof-K. At every step the post-encoder activations of a batch form a B×nB\times n matrix, nn the dictionary size, flattened into N=B⋅nN=B\cdot n entries of which the largest B⋅kB\cdot k are retained: one selection per step over a pool of 10810^{8} to 10910^{9} elements with retained fraction k/nk/n, on the critical path of training.

We train BatchTopK SAEs on residual-stream activations of GPT-2 XL [17] on a DGX H100 for 10910^{9} tokens, with dictionary sizes n∈{216,217,218}n\in\{2^{16},2^{17},2^{18}\} and batch sizes BB from 4,0964{,}096 to 12,28812{,}288, so that the pool N=B⋅nN=B\cdot n ranges from 2.7⋅1082.7\cdot 10^{8} to 2.1⋅1092.1\cdot 10^{9} elements (training details in Appendix D). The two runs of each configuration differ only in the selection call, torch.topk or Prof-K in exact mode, so the selected set and hence the training trajectory coincide, and the two runs are virtually identical in every tracked metric, as they should be. Table 1 reports the mean step time, the end-to-end speed-up and the GPU-hours of the full run. The speed-up grows with the pool size, from 1.20×1.20\times at 2.7⋅1082.7\cdot 10^{8} to 1.30×1.30\times at 2.1⋅1092.1\cdot 10^{9} elements, as the selection takes a growing share of the step, and it depends on NN alone: the configurations (217,8192)(2^{17},8192) and (218,4096)(2^{18},4096) with the same NN give the same gain, and the sparsity level does not enter, since top-kk latency is nearly flat in kk (Figure 2). In absolute terms Prof-K removes 0.60.6 to 2.12.1 GPU-hours from each 10910^{9}-token run.

Table 1: BatchTopK SAE training on GPT-2 XL activations (DGX H100, 10910^{9} tokens). N=B⋅nN=B\cdot n is the flattened pool size, ρ\rho is the share of the baseline step spent in top-kk, and the last two columns give the GPU-hours of the full run.
step time (ms) GPU-hours
nn BB NN torch.topk Prof-K ρ\rho speed-up torch.topk Prof-K
2162^{16} 4,0964{,}096 2.7⋅1082.7\cdot 10^{8} 50.3 41.9 18.5% 1.20×1.20\times 3.42 2.85
2172^{17} 4,0964{,}096 5.4⋅1085.4\cdot 10^{8} 71.4 58.4 22.4% 1.22×1.22\times 4.84 3.96
2172^{17} 6,1446{,}144 8.1⋅1088.1\cdot 10^{8} 101.8 81.4 23.6% 1.25×1.25\times 4.60 3.68
2172^{17} 8,1928{,}192 1.1⋅1091.1\cdot 10^{9} 132.2 104.4 24.2% 1.27×1.27\times 4.48 3.54
2172^{17} 12,28812{,}288 1.6⋅1091.6\cdot 10^{9} 193.0 150.4 24.9% 1.28×1.28\times 4.36 3.40
2182^{18} 4,0964{,}096 1.1⋅1091.1\cdot 10^{9} 142.8 115.0 22.4% 1.24×1.24\times 9.69 7.80
2182^{18} 6,1446{,}144 1.6⋅1091.6\cdot 10^{9} 203.6 161.0 23.6% 1.26×1.26\times 9.21 7.28
2182^{18} 8,1928{,}192 2.1⋅1092.1\cdot 10^{9} 264.4 203.5 25.1% 1.30×1.30\times 8.97 6.91

Application: Top-kk Gradient Sparsification

Top-kk gradient sparsification typically communicates only the largest 0.10.1 to 1%1\% of the gradient at every step and keeps the rest as an error-feedback residual [4]. For a model with roughly 10910^{9} parameters this is one selection over 10910^{9} elements per step, and its cost has long been the main objection to the method: the selection has been measured at the same order as the backward pass or as the communication it is meant to save [18, 19], and Yoon and Oh [20] report uncompressed training achieving 4.67×4.67\times higher throughput than exact Top-kk sparsification on eight A100s, with the selection as the dominant cost. Several widely used remedies trade exactness for speed: thresholds estimated from a small sample [4] or from a Gaussian fit of the gradient distribution [18], or thresholds or index sets periodically recomputed and reused across steps  [21, 22]. The exact primitive is still expensive: interpolating the large-NN H100 measurements reported in Table 4 to the 1.56⋅1091.56\cdot 10^{9} parameters of GPT-2 XL at 0.1%0.1\% density gives estimated selection latencies of about 4040 ms for torch.topk and 4.54.5 ms for Prof-K. Prof-K nevertheless returns the exact top-kk set at every step. Failures of the sampled filter are detected and handled, so the returned set remains exact. Exact per-step selection is therefore practical in this regime and avoids both the estimation risk of the threshold-based schemes, whose selected count can fall far short of kk and stall convergence when the gradient distribution departs from the assumed one [23], and the resampling interval of the index-reuse schemes, whose extremes degrade the loss [22]. Fine-tuning GPT-2 XL with error-feedback top-kk sparsification at 0.1%0.1\% density on a single H100 confirms this in the full training loop; Table 2 shows a 1.38×1.38\times step-time speed-up for GPT-2 XL and 1.16×1.16\times for GPT-2 Medium, with identical selected sets at every step.

Table 2: Single-H100 fine-tuning with error-feedback Top-kk gradient sparsification at 0.1%0.1\% density. Both selectors return the same Top-kk set at every step.
step time (ms)
Model Parameters torch.topk Prof-K speed-up
GPT-2 Medium 355355M 8585 7373 1.16×1.16\times
GPT-2 XL 1.561.56B 172172 125125 1.38×1.38\times

6 Conclusion

We introduced Prof-K, a probabilistic one-pass filtering method for exact top-kk selection on GPUs. It estimates a threshold from a small sample, streams once through the input into a compact candidate buffer, and selects the top-kk from this buffer, while a run-time check guarantees that every output is exact. A distribution-free analysis gives the failure probability exactly and yields the smallest parameters for a prescribed budget. On a DGX H100, Prof-K is up to 16×16\times faster than torch.topk and 3.1×3.1\times faster than RadiK, requires only a few megabytes of auxiliary memory, and scales to inputs that radix selection cannot process. At GPT-2 XL scale, it reduces the training time of BatchTopK sparse autoencoders and top-kk gradient sparsification, making it a practical building block for sparse machine learning workloads.

Limitations

Prof-K offers no advantage on small inputs, where the fixed cost of its four stages is comparable to that of a full exact selection and RadiK remains slightly faster unless many rows are batched. Moreover, the sample-size rule depends on a cost ratio calibrated once per GPU, and the method has not been evaluated on multi-GPU inputs, which we leave to future work.

References

  • [1] Reza Akbarinia, Esther Pacitti, and Patrick Valduriez. Best position algorithms for efficient top-k query processing. Information Systems, 36(6):973–989, 2011.
  • [2] Anil Shanbhag, Holger Pirk, and Samuel Madden. Efficient top-k query processing on massively parallel hardware. In Proceedings of the 2018 International Conference on Management of Data, pages 1557–1570, 2018.
  • [3] Bart Bussmann, Patrick Leask, and Neel Nanda. Batchtopk sparse autoencoders. arXiv preprint arXiv:2412.06410, 2024.
  • [4] Yujun Lin, Song Han, Huizi Mao, Yu Wang, and William J. Dally. Deep gradient compression: Reducing the communication bandwidth for distributed training. In International Conference on Learning Representations, 2018. URL https://openreview.net/forum?id=SkhQHMW0W.
  • [5] Noam Shazeer, Azalia Mirhoseini, Krzysztof Maziarz, Andy Davis, Quoc Le, Geoffrey Hinton, and Jeff Dean. Outrageously large neural networks: The sparsely-gated mixture-of-experts layer. In International Conference on Learning Representations, 2017. URL https://openreview.net/forum?id=B1ckMDqlg.
  • [6] Dmitry Lepikhin, HyoukJoong Lee, Yuanzhong Xu, Dehao Chen, Orhan Firat, Yanping Huang, Maxim Krikun, Noam Shazeer, and Zhifeng Chen. Gshard: Scaling giant models with conditional computation and automatic sharding. In International Conference on Learning Representations, 2021. URL https://openreview.net/forum?id=qrwe7XHTmYb.
  • [7] William Fedus, Barret Zoph, and Noam Shazeer. Switch transformers: Scaling to trillion parameter models with simple and efficient sparsity. Journal of Machine Learning Research, 23(120):1–39, 2022.
  • [8] Tolu Alabi, Jeffrey D Blanchard, Bradley Gordon, and Russel Steinbach. Fast k-selection algorithms for graphics processing units. Journal of Experimental Algorithmics (JEA), 17:4.1–4.29, 2012.
  • [9] Anil Gaihre, Da Zheng, Scott Weitze, Lingda Li, Shuaiwen Leon Song, Caiwen Ding, Xiaoye S Li, and Hang Liu. Dr. top-k: delegate-centric top-k on gpus. In Proceedings of the International Conference for High Performance Computing, Networking, Storage and Analysis, pages 1–14, 2021.
  • [10] Jingrong Zhang, Akira Naruse, Xipeng Li, and Yong Wang. Parallel top-k algorithms on gpu: A comprehensive study and new methods. In Proceedings of the International Conference for High Performance Computing, Networking, Storage and Analysis, pages 1–13, 2023.
  • [11] Xi Xie, Yuebo Luo, Hongwu Peng, and Caiwen Ding. Rtop-k: Ultra-fast row-wise top-k selection for neural network acceleration on gpus. In International Conference on Learning Representations, 2025. URL https://proceedings.iclr.cc/paper_files/paper/2025/file/ca1b93fc0f3560ba84eb0bc8de6d8f91-Paper-Conference.pdf.
  • [12] Yifei Li, Bole Zhou, Jiejing Zhang, Xuechao Wei, Yinghan Li, and Yingda Chen. Radik: scalable and optimized gpu-parallel radix top-k selection. In Proceedings of the 38th ACM International Conference on Supercomputing, pages 537–548, 2024.
  • [13] Felix Chern, Blake Hechtman, Andy Davis, Ruiqi Guo, David Majnemer, and Sanjiv Kumar. Tpu-knn: K nearest neighbor search at peak flop/s. In Advances in Neural Information Processing Systems, volume 35, pages 15489–15501, 2022.
  • [14] Bart Bussmann, Noa Nabeshima, Adam Karvonen, and Neel Nanda. Learning multi-level features with Matryoshka sparse autoencoders. In Proceedings of the 42nd International Conference on Machine Learning, volume 267 of Proceedings of Machine Learning Research, pages 6077–6101, 2025.
  • [15] Julian Minder, Clément Dumas, Caden Juang, Bilal Chughtai, and Neel Nanda. Overcoming sparsity artifacts in crosscoders to interpret chat-tuning. In Advances in Neural Information Processing Systems, volume 38, pages 106423–106474, 2025.
  • [16] Michael Greenwald and Sanjeev Khanna. Space-efficient online computation of quantile summaries. ACM SIGMOD Record, 30(2):58–66, 2001.
  • [17] Alec Radford, Jeffrey Wu, Rewon Child, David Luan, Dario Amodei, and Ilya Sutskever. Language models are unsupervised multitask learners. OpenAI blog, 2019.
  • [18] Shaohuai Shi, Xiaowen Chu, Ka Chun Cheung, and Simon See. Understanding top-k sparsification in distributed deep learning. arXiv preprint arXiv:1911.08772, 2019.
  • [19] Saurabh Agarwal, Hongyi Wang, Shivaram Venkataraman, and Dimitris Papailiopoulos. On the utility of gradient compression in distributed training systems. In Proceedings of Machine Learning and Systems, volume 4, pages 652–672, 2022.
  • [20] Daegun Yoon and Sangyoon Oh. Empirical analysis on top-k gradient sparsification for distributed deep learning in a supercomputing environment. arXiv preprint arXiv:2209.08497, 2022.
  • [21] Shigang Li and Torsten Hoefler. Near-optimal sparse allreduce for distributed deep learning. In Proceedings of the 27th ACM SIGPLAN Symposium on Principles and Practice of Parallel Programming, pages 135–149, 2022.
  • [22] Mingkai Zheng and Zhao Zhang. Radius: Range-based gradient sparsity for large foundation model pre-training. In Proceedings of Machine Learning and Systems, volume 7, 2025.
  • [23] Ahmed M. Abdelmoniem, Ahmed Elzanaty, Mohamed-Slim Alouini, and Marco Canini. An efficient statistical-based gradient compression technique for distributed training systems. In Proceedings of Machine Learning and Systems, volume 3, pages 297–322, 2021.
  • [24] Peter J. Acklam. An algorithm for computing the inverse normal cumulative distribution function. Archived copy: https://web.archive.org/web/20151030215612/http://home.online.no/~pjacklam/notes/invnorm/.
  • [25] Edmund A Cornish and Ronald A Fisher. Moments and cumulants in the specification of distributions. Revue de l’Institut International de Statistique, 5(4):307–320, 1938.
  • [26] Edwin B. Wilson and Margaret M. Hilferty. The distribution of chi-square. Proceedings of the National Academy of Sciences, 17(12):684–688, 1931.

Appendix A Proofs

Proof of Lemma 1.

All (NS)\binom{N}{S} subsets of size SS are equally likely, and exactly (Kj)​(N−KS−j)\binom{K}{j}\binom{N-K}{S-j} of them contain precisely jj indices of TKT_{K}, which gives Equation 1 for XKX_{K}; Equation 4 follows from Equation 3. Monotonicity in tt is immediate from the definition of a cumulative distribution function. For K<K′K<K^{\prime} we have TK⊂TK′T_{K}\subset T_{K^{\prime}}, hence XK≤XK′X_{K}\leq X_{K^{\prime}} pointwise and Pr⁡(XK′≤t−1)≤Pr⁡(XK≤t−1)\Pr(X_{K^{\prime}}\leq t-1)\leq\Pr(X_{K}\leq t-1). ∎

Proof of Proposition 2.

The event {R<k}\{R<k\} is the event that τ\tau is one of the k−1k-1 largest entries of xx. Since τ\tau is the tt-th largest sampled value, this happens if and only if the tt largest sampled values all belong to Tk−1T_{k-1}, that is, if and only if Xk−1≥tX_{k-1}\geq t. Hence Pr⁡(R<k)=Pr⁡(Xk−1≥t)=1−Pr⁡(Xk−1≤t−1)=1−G⁡(k−1)\Pr(R<k)=\Pr(X_{k-1}\geq t)=1-\Pr(X_{k-1}\leq t-1)=1-G(k-1), and Lemma 1 gives the second equality. ∎

Proof of Proposition 3.

The event {R>M}\{R>M\} is the event that τ\tau is not among the MM largest entries of xx, which happens if and only if fewer than tt sampled indices belong to TMT_{M}, that is, XM≤t−1X_{M}\leq t-1. Hence Pr⁡(R>M)=G⁡(M)\Pr(R>M)=G(M), and Lemma 1 gives the second equality. ∎

Proof of Theorem 4.

Write F=FN,k−1,SF=F_{N,k-1,S}. By Proposition 2, Pr⁡(R<k)≤εA\Pr(R<k)\leq\varepsilon_{A} holds if and only if F⁡(t−1)≥1−εAF(t-1)\geq 1-\varepsilon_{A}. Since FF is nondecreasing, the set of such tt is {t:t−1≥F−1​(1−εA)}\{t:t-1\geq F^{-1}(1-\varepsilon_{A})\}, whose smallest element is t⋆t^{\star}; this proves (i). The quantile F−1​(1−εA)F^{-1}(1-\varepsilon_{A}) is at most SS, and it equals SS only if F⁡(S−1)<1−εAF(S-1)<1-\varepsilon_{A}, that is, Pr⁡(Xk−1=S)>εA\Pr(X_{k-1}=S)>\varepsilon_{A}, which the assumption excludes; hence t⋆≤St^{\star}\leq S. For (ii), by Proposition 3 the condition Pr⁡(R>M)≤εB\Pr(R>M)\leq\varepsilon_{B} reads G⁡(M)≤εBG(M)\leq\varepsilon_{B} with t=t⋆t=t^{\star}. By Lemma 1, GG is nonincreasing in MM, so the set of M∈{k,…,N}M\in\{k,\dots,N\} satisfying it is an interval ending at NN; it is nonempty because XN=S≥t⋆X_{N}=S\geq t^{\star} gives G⁡(N)=0G(N)=0. Its smallest element is M⋆M^{\star} by definition. Finally, AA and BB are disjoint because M⋆≥kM^{\star}\geq k, which gives Equation 8, and the bound εA+εB\varepsilon_{A}+\varepsilon_{B} follows from (i) and (ii). ∎

Proof of Corollary 5.

t≥t⋆t\geq t^{\star} is (i) of Theorem 4. Since GG is nondecreasing in tt by Lemma 1, G⁡(N,M,S,t⋆)≤G⁡(N,M,S,t)≤εBG(N,M,S,t^{\star})\leq G(N,M,S,t)\leq\varepsilon_{B}, so MM satisfies the defining condition of M⋆M^{\star} and M≥M⋆M\geq M^{\star} by minimality. ∎

Proof of Proposition 7.

C~\tilde{C} is a sum of a linear function and two strictly convex decreasing functions of SS, so it is strictly convex, and C~​(S)→∞\tilde{C}(S)\to\infty as S→0S\to 0 or S→∞S\to\infty, so the minimizer exists and is unique. It is the root of C~′​(S)=0\tilde{C}^{\prime}(S)=0, that is, of r=f1​(S)+f2​(S)r=f_{1}(S)+f_{2}(S) with f1(S)=zk​NS−3/2=r(S1/S)3/2f_{1}(S)=z\sqrt{kN}\,S^{-3/2}=r(S_{1}/S)^{3/2} and f2​(S)=κ​N​S−2=r​(S2/S)2f_{2}(S)=\kappa NS^{-2}=r(S_{2}/S)^{2}, both decreasing. At S=max⁡(S1,S2)S=\max(S_{1},S_{2}) one of f1,f2f_{1},f_{2} equals rr, so f1+f2≥rf_{1}+f_{2}\geq r and S∘≥max⁡(S1,S2)S^{\circ}\geq\max(S_{1},S_{2}). At S=22/3​max⁡(S1,S2)S=2^{2/3}\max(S_{1},S_{2}) we have f1≤r/2f_{1}\leq r/2 and f2≤2−4/3r≤r/2f_{2}\leq 2^{-4/3}r\leq r/2, so f1+f2≤rf_{1}+f_{2}\leq r and S∘≤22/3​max⁡(S1,S2)S^{\circ}\leq 2^{2/3}\max(S_{1},S_{2}). ∎

Proof of Lemma 6.

Write the sampled indices in increasing order of rank as ρ1<ρ2<⋯<ρS\rho_{1}<\rho_{2}<\dots<\rho_{S}, so that ρt=R\rho_{t}=R. The SS sampled ranks split the N−SN-S unsampled ranks into S+1S+1 gaps of sizes D1,…,DS+1≥0D_{1},\dots,D_{S+1}\geq 0, where DiD_{i} is the number of unsampled ranks strictly between ρi−1\rho_{i-1} and ρi\rho_{i} (with ρ0=0\rho_{0}=0 and ρS+1=N+1\rho_{S+1}=N+1), and ρt=t+∑i≤tDi\rho_{t}=t+\sum_{i\leq t}D_{i}. Since 𝒮\mathcal{S} is a uniformly random subset of size SS, the vector (D1,…,DS+1)(D_{1},\dots,D_{S+1}) is uniformly distributed over the compositions of N−SN-S into S+1S+1 nonnegative parts, a distribution that is invariant under permutations of the parts; hence the DiD_{i} are exchangeable and 𝔼⁡[Di]=(N−S)/(S+1)\E[D_{i}]=(N-S)/(S+1) for every ii. Therefore

𝔼⁡[R]=t+t​N−SS+1=N+1S+1​t.∎\E[R]=t+t\,\frac{N-S}{S+1}=\frac{N+1}{S+1}\,t.\qed

Appendix B Fast Parameter Computation

This appendix derives the closed-form surrogates for t⋆t^{\star} and M⋆M^{\star} that Algorithm 1 in Section 4 uses, derives the sample-size rule of Section 4, and compares the surrogates with the exact parameters of Theorem 4. Throughout, εA=εB=ε/2\varepsilon_{A}=\varepsilon_{B}=\varepsilon/2 and z=Φ−1​(1−ε/2)z=\Phi^{-1}(1-\varepsilon/2), where Φ\Phi is the standard normal distribution function; Φ−1\Phi^{-1} is itself evaluated by the rational approximation of Acklam [24], whose relative error is below 1.15⋅10−91.15\cdot 10^{-9} over the whole domain, so that the fast path has no library dependence. Each surrogate is chosen according to the regime in which it is accurate, and each is designed to err on the side of a larger tt or a larger MM, since an underestimate of either would violate its budget.

Threshold rank. By Theorem 4, t⋆−1t^{\star}-1 is the (1−εA)(1-\varepsilon_{A})-quantile of Xk−1∼Hypergeometric⁡(N,k−1,S)X_{k-1}\sim\Hypergeometric(N,k-1,S), whose mean and variance are

μ=(k−1)​SN,σ2=μ⁡(1−k−1N)​N−SN−1.\mu=\frac{(k-1)S}{N},\qquad\sigma^{2}=\mu\Bigl(1-\frac{k-1}{N}\Bigr)\frac{N-S}{N-1}. (9)

The quantity μ\mu, the expected number of sampled indices among the top-(k−1)(k-1), determines the shape of the distribution and hence the appropriate approximation. When μ≥15\mu\geq 15, Xk−1X_{k-1} is close to normal, but it remains right-skewed, and the symmetric normal quantile μ+z​σ\mu+z\sigma systematically underestimates its upper tail. We therefore use the second-order Cornish–Fisher expansion [25], which corrects the normal quantile of a distribution with skewness γ\gamma by the term (z2−1)​γ​σ/6(z^{2}-1)\gamma\sigma/6. For the hypergeometric distribution γ​σ=(N−2​(k−1))​(N−2​S)/(N⁡(N−2))≤1\gamma\sigma=(N-2(k-1))(N-2S)/(N(N-2))\leq 1, with equality approached in the sparse regime k≪Nk\ll N, S≪NS\ll N; replacing γ​σ\gamma\sigma by its upper bound 11 keeps the correction conservative. This gives

t=⌈μ+z​σ+z2−16⌉+1(μ≥15).t=\Bigl\lceil\mu+z\sigma+\frac{z^{2}-1}{6}\Bigr\rceil+1\qquad(\mu\geq 15). (10)

When μ<15\mu<15, the normal approximation is unreliable, but in this regime k−1≪Nk-1\ll N and Xk−1X_{k-1} is a count of rare events, for which Poisson⁡(μ)\mathrm{Poisson}(\mu) is the natural limit: Hypergeometric⁡(N,k−1,S)\Hypergeometric(N,k-1,S) converges to Poisson⁡(μ)\mathrm{Poisson}(\mu) as N→∞N\to\infty with (k−1)​S/N=μ(k-1)S/N=\mu fixed. Since Poisson⁡(μ)\mathrm{Poisson}(\mu) has variance μ≥σ2\mu\geq\sigma^{2}, its upper quantiles dominate those of Xk−1X_{k-1}, and replacing one by the other is again conservative. We therefore take t=1+j⋆t=1+j^{\star} with

j⋆=min⁡{j≥0:∑i=0je−μ​μii!≥1−εA}(μ<15),j^{\star}=\min\Bigl\{\,j\geq 0:\sum_{i=0}^{j}e^{-\mu}\frac{\mu^{i}}{i!}\geq 1-\varepsilon_{A}\Bigr\}\qquad(\mu<15), (11)

evaluated by accumulating the terms through the recurrence p0=e−μp_{0}=e^{-\mu}, pi=pi−1​μ/ip_{i}=p_{i-1}\mu/i, which stops after O⁡(μ+z​μ)O(\mu+z\sqrt{\mu}) iterations, that is, a few dozen at most. For k=1k=1 this gives μ=0\mu=0 and t=1t=1, which is exact: the sample maximum never exceeds the global maximum, so Pr⁡(R<1)=0\Pr(R<1)=0.

Buffer capacity. By Proposition 3, M⋆M^{\star} is the smallest MM with Pr⁡(R>M)≤εB\Pr(R>M)\leq\varepsilon_{B}, that is, the (1−εB)(1-\varepsilon_{B})-quantile of the rank RR itself. Its distribution is that of the tt-th smallest of SS positions drawn uniformly without replacement from [N][N], and its mean is known exactly.

Lemma 6.

For 1≤t≤S≤N1\leq t\leq S\leq N, 𝔼⁡[R]=N+1S+1​t\E[R]=\dfrac{N+1}{S+1}\,t.

As N→∞N\to\infty with SS fixed, the sampled positions become independent uniform points, and R/(N+1)R/(N+1) converges in distribution to the tt-th order statistic of SS independent Uniform⁡(0,1)\mathrm{Uniform}(0,1) variables; as S→∞S\to\infty with tt fixed, this order statistic rescaled by S+1S+1 converges to Gamma⁡(t,1)\mathrm{Gamma}(t,1). We therefore approximate RR by N+1S+1​Γt\frac{N+1}{S+1}\Gamma_{t} with Γt∼Gamma⁡(t,1)\Gamma_{t}\sim\mathrm{Gamma}(t,1), a scaling that matches 𝔼⁡[R]\E[R] exactly by Lemma 6. This choice retains the feature that matters most for an upper quantile: for small tt the distribution of RR is strongly right-skewed, and a normal approximation with the same mean and variance would underestimate M⋆M^{\star} severely, whereas Gamma⁡(t,1)\mathrm{Gamma}(t,1) has the same skew. To evaluate the quantile of Γt\Gamma_{t} in closed form we use the Wilson–Hilferty transform [26], by which (Γt/t)1/3(\Gamma_{t}/t)^{1/3} is approximately normal with mean 1−h1-h and variance hh, where h=1/(9​t)h=1/(9t). This yields

M=⌈N+1S+1​t​(1−h+z​h)3⌉,h=19​t,M=\Bigl\lceil\frac{N+1}{S+1}\,t\,\bigl(1-h+z\sqrt{h}\bigr)^{3}\Bigr\rceil,\qquad h=\frac{1}{9t}, (12)

for every tt. For large tt the cube expands to t+z​t+O⁡(1)t+z\sqrt{t}+O(1), the normal quantile of Gamma⁡(t,1)\mathrm{Gamma}(t,1), so Equation 12 reduces to the normal approximation exactly where the latter is valid. At the other extreme t=1t=1, where Γ1\Gamma_{1} is exponential with (1−εB)(1-\varepsilon_{B})-quantile ln⁡(1/εB)\ln(1/\varepsilon_{B}), Equation 12 overestimates the quantile by about 3%3\% for εB=5×10−4\varepsilon_{B}=5\times 10^{-4}, which is the conservative direction. The complete procedure, combining Equations 10, 11 and 12 with the choice of SS derived next, is summarized in Algorithm 1.

Choice of the Sample Size

Theorem 4 fixes t⋆t^{\star} and M⋆M^{\star} for a given SS but leaves SS free, and SS controls a trade-off: a larger sample makes the threshold more accurate and the buffer smaller, at the price of more sampling work. We choose SS by minimizing a cost model of the stages of Prof-K on a GPU. The filtering pass streams the NN elements once at memory bandwidth and does not depend on SS or MM. The other stages process a number of elements proportional to SS or MM: drawing the sample is a random gather of SS elements at a per-element cost gg, well above the streaming cost since each random access fetches a full memory sector; the threshold is a top-tt selection over the SS sampled values at a per-element cost bb; and the refinement is a top-kk selection over the MM buffer slots at a per-element cost d⁡(k)d(k) that grows with kk, since selecting many elements from the buffer is closer to a sort than to a scan. Fixed launch overheads do not affect the minimizer. The overhead beyond the mandatory scan is therefore

C⁡(S)=(g+b)​S+d⁡(k)​M⋆​(S),C(S)=(g+b)\,S+d(k)\,M^{\star}(S), (13)

where M⋆​(S)M^{\star}(S) denotes the exact buffer capacity of Theorem 4 at sample size SS. To minimize Equation 13 in closed form we use the leading-order behavior of M⋆​(S)M^{\star}(S), which follows from the surrogates above:

M⋆​(S)≈k+2​z​k​NS+κ​NS,ln⁡(1/εB)≤κ≤z2+12,M^{\star}(S)\;\approx\;k+2z\sqrt{\frac{kN}{S}}+\kappa\,\frac{N}{S},\qquad\ln(1/\varepsilon_{B})\leq\kappa\leq z^{2}+\tfrac{1}{2}, (14)

where the lower end of the range for κ\kappa is attained at t=1t=1 and the upper end for large tt; we use κ=z2+12\kappa=z^{2}+\frac{1}{2}. The square-root term is the familiar sampling error of a quantile estimate; the N/SN/S term is the contribution of the skew of RR at small tt, and it dominates in the sparse regime μ≲1\mu\lesssim 1, where the buffer is governed by N/SN/S rather than by kk. Substituting Equation 14 into Equation 13 and writing r=(g+b)/d⁡(k)r=(g+b)/d(k) for the cost of sampling one element relative to selecting one from the buffer, the model has a unique minimizer, which we characterize as follows.

Proposition 7 (Sample size).

Let r,z,κ>0r,z,\kappa>0 and let C~​(S)=r​S+2​z​k​N/S+κ​N/S\tilde{C}(S)=rS+2z\sqrt{kN/S}+\kappa N/S for S>0S>0. Then C~\tilde{C} is strictly convex with a unique minimizer S∘S^{\circ}, and with

S1=(zr)2/3​(k​N)1/3,S2=(κ​Nr)1/2,S_{1}=\Bigl(\frac{z}{r}\Bigr)^{2/3}(kN)^{1/3},\qquad S_{2}=\Bigl(\frac{\kappa N}{r}\Bigr)^{1/2}, (15)

it satisfies max⁡(S1,S2)≤S∘≤22/3​max⁡(S1,S2)\max(S_{1},S_{2})\leq S^{\circ}\leq 2^{2/3}\max(S_{1},S_{2}).

The two candidates correspond to the two terms of Equation 14: for k≳z​r​Nk\gtrsim z\sqrt{rN} the optimal sample size scales as (k​N)1/3(kN)^{1/3}, and below that as N\sqrt{N}, independently of kk, because the buffer is then governed by N/SN/S. We use S=clip⁡(max⁡(S1,S2),Smin,Smax)S=\mathrm{clip}(\max(S_{1},S_{2}),S_{\min},S_{\max}), which by Proposition 7 is within a factor 22/3≈1.592^{2/3}\approx 1.59 of the model optimum, and the sensitivity study in Appendix E.1 shows that the latency is flat over a considerably wider range of SS around it. The lower bound SminS_{\min} reflects the fixed launch cost of the selection routine, below which a smaller sample saves nothing, and SmaxS_{\max} keeps the sampling stage negligible against the scan. The only hardware-dependent quantity is the ratio rr, which we calibrate once per GPU from a sweep over SS (Appendix E.1). The calibration confirms the two regimes the model predicts: for kk up to a few thousand the buffer selection costs less per element than the sampling, rr is large, and the rule returns the floor SminS_{\min}; for larger kk the buffer selection dominates, 1/r1/r grows linearly in log2⁡k\log_{2}k, and the sample grows with kk and NN. When the retained fraction exceeds k/N≈2−10k/N\approx 2^{-10} the threshold selection over the sample, whose rank tt grows with S​k/NSk/N, becomes the dominant cost as well, and the floor is again the best choice.

Accuracy

Table 3 compares the parameters of Algorithm 1 with the exact optimal ones of Theorem 4 at ε=10−3\varepsilon=10^{-3} over 2121 configurations spanning NN from 2202^{20} to 2312^{31} and kk from 252^{5} to 2202^{20}, with the sample size of Algorithm 1 in both cases. For each pair of parameters the table also reports the exact failure probability Pr⁡(R<k)+Pr⁡(R>M)\Pr(R<k)+\Pr(R>M), evaluated with the hypergeometric distribution function of Propositions 2 and 3. The surrogate threshold rank equals t⋆t^{\star} in 1919 configurations and exceeds it by one in the other two; the surrogate buffer capacity exceeds M⋆M^{\star} by 0.10.1 to 3%3\%, with the largest excess at t=1t=1, where the Wilson–Hilferty quantile is least accurate, and the excess vanishes as tt grows. The attained failure probability is below the budget in every configuration for both parameter sets, and below that of the exact parameters for the surrogates, which confirms that the approximations err on the conservative side.

Table 3: Fast approximate parameters (Algorithm 1) against the exact optimal parameters (Theorem 4) at ε=10−3\varepsilon=10^{-3}, with the exact failure probability Pr⁡(R<k)+Pr⁡(R>M)\Pr(R<k)+\Pr(R>M) attained by each pair.
NN kk SS t⋆t^{\star} tt M⋆M^{\star} MM M/M⋆M/M^{\star} Pr⁡(fail)\Pr(\text{fail}), exact Pr⁡(fail)\Pr(\text{fail}), approx.
2202^{20} 252^{5} 8,192 4 4 1,777 1,801 1.014 5.99⋅10−45.99\cdot 10^{-4} 5.28⋅10−45.28\cdot 10^{-4}
2202^{20} 282^{8} 8,192 9 9 2,835 2,855 1.007 7.09⋅10−47.09\cdot 10^{-4} 6.60⋅10−46.60\cdot 10^{-4}
2202^{20} 2112^{11} 8,192 32 32 6,883 6,909 1.004 7.53⋅10−47.53\cdot 10^{-4} 7.09⋅10−47.09\cdot 10^{-4}
2202^{20} 2142^{14} 8,192 167 168 27,141 27,385 1.009 9.70⋅10−49.70\cdot 10^{-4} 7.64⋅10−47.64\cdot 10^{-4}
2242^{24} 252^{5} 16,384 2 2 10,232 10,422 1.019 9.35⋅10−49.35\cdot 10^{-4} 8.57⋅10−48.57\cdot 10^{-4}
2242^{24} 282^{8} 16,384 4 4 14,259 14,402 1.010 6.28⋅10−46.28\cdot 10^{-4} 5.75⋅10−45.75\cdot 10^{-4}
2242^{24} 2112^{11} 20,650 10 10 19,281 19,362 1.004 7.91⋅10−47.91\cdot 10^{-4} 7.60⋅10−47.60\cdot 10^{-4}
2242^{24} 2142^{14} 36,403 58 58 39,777 39,822 1.001 8.27⋅10−48.27\cdot 10^{-4} 8.09⋅10−48.09\cdot 10^{-4}
2242^{24} 2172^{17} 16,384 168 168 218,748 219,092 1.002 8.87⋅10−48.87\cdot 10^{-4} 8.49⋅10−48.49\cdot 10^{-4}
2282^{28} 252^{5} 16,384 2 2 163,769 166,737 1.018 5.02⋅10−45.02\cdot 10^{-4} 4.26⋅10−44.26\cdot 10^{-4}
2282^{28} 282^{8} 37,480 3 3 86,297 87,385 1.013 5.07⋅10−45.07\cdot 10^{-4} 4.47⋅10−44.47\cdot 10^{-4}
2282^{28} 2112^{11} 82,603 5 5 51,044 51,435 1.008 9.89⋅10−49.89\cdot 10^{-4} 9.45⋅10−49.45\cdot 10^{-4}
2282^{28} 2142^{14} 110,643 18 18 85,613 85,777 1.002 7.37⋅10−47.37\cdot 10^{-4} 7.19⋅10−47.19\cdot 10^{-4}
2282^{28} 2172^{17} 234,231 152 153 224,429 225,793 1.006 9.45⋅10−49.45\cdot 10^{-4} 8.20⋅10−48.20\cdot 10^{-4}
2282^{28} 2202^{20} 16,384 93 93 2,095,476 2,098,101 1.001 8.83⋅10−48.83\cdot 10^{-4} 8.60⋅10−48.60\cdot 10^{-4}
2312^{31} 252^{5} 16,384 1 1 996,031 1,026,231 1.030 7.36⋅10−47.36\cdot 10^{-4} 6.34⋅10−46.34\cdot 10^{-4}
2312^{31} 282^{8} 106,011 2 2 202,532 206,164 1.018 5.78⋅10−45.78\cdot 10^{-4} 5.03⋅10−45.03\cdot 10^{-4}
2312^{31} 2112^{11} 233,638 4 4 128,067 129,279 1.009 5.86⋅10−45.86\cdot 10^{-4} 5.35⋅10−45.35\cdot 10^{-4}
2312^{31} 2142^{14} 262,144 9 9 181,989 182,719 1.004 7.37⋅10−47.37\cdot 10^{-4} 7.08⋅10−47.08\cdot 10^{-4}
2312^{31} 2172^{17} 262,144 32 32 441,824 442,225 1.001 7.76⋅10−47.76\cdot 10^{-4} 7.65⋅10−47.65\cdot 10^{-4}
2312^{31} 2202^{20} 262,144 168 168 1,752,499 1,752,830 1.000 9.09⋅10−49.09\cdot 10^{-4} 9.04⋅10−49.04\cdot 10^{-4}

Appendix C Implementation Details

The Filter Kernel

The only custom kernel in Prof-K is the filtering pass, shown in ; sampling, the threshold selection over the sample and the final selection over the buffer are standard library calls (torch.topk on the SS sampled values and on the MM buffer slots). The kernel is written in Triton and launched on a two-dimensional grid, one axis over blocks of BLOCK_SIZE consecutive elements of a row and one over the rows of the batch, so that batched inputs are handled by the same code with a per-row threshold, buffer and counter. Each program loads its block once, compares it against the threshold of its row, and counts the elements that pass. If any do, it reserves a contiguous range of buffer slots for them with a single atomic addition on the per-row counter, computes the position of each retained element within that range by an in-block prefix sum, and writes values and indices to the reserved slots. The pass therefore streams the input exactly once, issues one atomic per block rather than one per retained element, and writes only the retained elements, which is what keeps it at the memory bandwidth of the device.

The counter is incremented by the full number of retained elements in the block, whereas the stores are masked to the buffer capacity. After the pass the counter of each row therefore holds the exact rank RR of its threshold, whether or not the buffer overflowed, and the run-time check k≤R≤Mk\leq R\leq M of Section 3 is a comparison of BB integers, without any extra pass over the input. The block size and the number of warps are selected by Triton’s autotuner from the configurations listed, keyed on the row length, and the counters are reset to zero between autotuning trials so that the trials do not accumulate into them.

1 import triton
2 import triton.language as tl
3
4
5 @triton.autotune(
6 configs=[
7 triton.Config({’BLOCK_SIZE’: 1024}, num_warps=8, num_stages=3),
8 triton.Config({’BLOCK_SIZE’: 2048}, num_warps=8, num_stages=4),
9 triton.Config({’BLOCK_SIZE’: 4096}, num_warps=8, num_stages=4),
10 ],
11 key=[’n_elements’],
12 reset_to_zero=[’count_ptr’],
13 )
14 @triton.jit
15 def _filter_topk_batched_kernel(
16 x_ptr, out_vals_ptr, out_idx_ptr, count_ptr, thresholds_ptr,
17 stride_xb: tl.int64, stride_ob: tl.int64, n_elements: tl.int64, buffer_cap,
18 BLOCK_SIZE: tl.constexpr,
19 ):
20 pid_chunk = tl.program_id(axis=0) # block of the input row
21 pid_batch = tl.program_id(axis=1) # row of the batch
22
23 # 64-bit offsets: inputs may exceed 2^31 elements
24 block_start = pid_chunk.to(tl.int64) * BLOCK_SIZE
25 offsets = block_start + tl.arange(0, BLOCK_SIZE)
26 read_mask = offsets < n_elements
27
28 x_offset = pid_batch.to(tl.int64) * stride_xb + offsets
29 x = tl.load(x_ptr + x_offset, mask=read_mask, other=-float(’inf’))
30 thresh = tl.load(thresholds_ptr + pid_batch)
31
32 pass_mask = (x >= thresh) & read_mask
33 pass_int = pass_mask.to(tl.int32)
34 n_pass = tl.sum(pass_int)
35
36 if n_pass > 0:
37 # one atomic per block reserves n_pass slots and counts every
38 # retained element, also beyond the buffer capacity
39 global_offset = tl.atomic_add(count_ptr + pid_batch, n_pass)
40 local_offset = tl.cumsum(pass_int) - 1
41 write_idx = global_offset + local_offset
42 write_mask = pass_mask & (write_idx < buffer_cap)
43
44 out_offset = pid_batch * stride_ob + write_idx
45 tl.store(out_vals_ptr + out_offset, x, mask=write_mask)
46 tl.store(out_idx_ptr + out_offset, offsets.to(tl.int64),
47 mask=write_mask)
Listing 1: The Triton filter kernel of Prof-K.

Appendix D Experimental Details and Additional Results

Latency of Every Configuration

Figure 5 reports the absolute latency of the three methods on every (N,k)(N,k) configuration of the grid on both GPUs, grouped by NN with kk increasing within each group.

Refer to caption
Figure 5: Latency (ms, log scale) of torch.topk, RadiK and Prof-K on every (N,k)(N,k) configuration of the grid, on the DGX H100 (top, k∈{25,27,…,219}k\in\{2^{5},2^{7},\dots,2^{19}\}) and the RTX 3060 (bottom, k∈{25,27,…,217}k\in\{2^{5},2^{7},\dots,2^{17}\}), at ε=10−3\varepsilon=10^{-3}. Groups are ordered by NN and, within a group, by kk.

Speed-up as a Function of NN

Figure 6 shows every (N,k)(N,k) configuration of the grid of Section 5 as one point, with the spread over kk made visible. Table 4 lists the configurations beyond N=230N=2^{30} on the H100, where RadiK cannot run because its 3232-bit indexing stops at 2312^{31} elements, so Prof-K is compared with torch.topk alone.

Figure 6: Speed-up of Prof-K as a function of NN; each point is one (N,k)(N,k) configuration, horizontally spread by kk, and the lines join the medians over kk. The dashed line marks parity.
Table 4: Prof-K against torch.topk on the DGX H100 for inputs beyond the reach of RadiK. Latency in ms, auxiliary memory in MB.
NN kk Prof-K (ms) torch.topk (ms) speed-up Prof-K (MB) torch.topk (MB)
2312^{31} 2102^{10} 3.14 53.8 17.1×17.1\times 3.1 65.0
2312^{31} 2142^{14} 3.19 53.5 16.8×16.8\times 5.8 65.2
2312^{31} 2182^{18} 3.56 53.4 15.0×15.0\times 24.1 68.0
2312^{31} 2222^{22} 7.25 54.0 7.5×7.5\times 238 177
3⋅1093\cdot 10^{9} 2102^{10} 4.67 74.6 16.0×16.0\times 5.4 91.4
3⋅1093\cdot 10^{9} 2182^{18} 5.61 74.6 13.3×13.3\times 32.4 94.4
4.3⋅1094.3\cdot 10^{9} 2102^{10} 6.61 106.7 16.1×16.1\times 5.7 130.0
4.3⋅1094.3\cdot 10^{9} 2182^{18} 7.51 106.8 14.2×14.2\times 35.1 133.0
4.3⋅1094.3\cdot 10^{9} 2222^{22} 15.4 106.5 6.9×6.9\times 285 178

Cost of the Run-Time Check

Exact mode is guaranteed to return the true top-kk; the price is the host synchronisation needed to inspect the count RR after the filtering pass, and the rare redraw. On the H100 this costs between 5%5\% of the latency at N=230N=2^{30} and 40%40\% at N=221N=2^{21}, where the synchronisation exposes kernel-launch latency that is otherwise hidden. Omitting the check, which is appropriate when an occasional miss is acceptable, recovers this overhead, and we study the resulting accuracy–speed trade-off in Appendix E.

Appendix E Ablations

E.1 Sample Size

The sample size SS is the one parameter of Prof-K that the exact analysis of Section 4 leaves free, and Proposition 7 chooses it from a cost model whose single hardware constant, the cost ratio rr, has to be calibrated. We therefore sweep SS on the RTX 3060 to calibrate rr and to measure how far the resulting rule lands from the latency-optimal sample size. The sweep covers S∈{212,213,…,218}S\in\{2^{12},2^{13},\dots,2^{18}\}, capped at N/32N/32, for every N∈{221,…,228}N\in\{2^{21},\dots,2^{28}\} and k∈{25,28,211,214,217}k\in\{2^{5},2^{8},2^{11},2^{14},2^{17}\}; for each SS the threshold rank and buffer capacity are the exact optimal ones of Theorem 4 at ε=10−3\varepsilon=10^{-3}, and the latency is that of exact mode, averaged over 88 input tensors and 2020 timed calls each after a warm-up.

Calibration

Fitting the cost model C=c+(g+b)​S+d⁡(k)​M⋆​(S)C=c+(g+b)S+d(k)M^{\star}(S) to the 40×540\times 5 measurements with S≥214S\geq 2^{14}, with one intercept cc per configuration, one sampling cost g+bg+b, and one buffer cost d⁡(k)d(k) per value of kk, explains the latencies to within 4.6%4.6\% on average (R2=0.998R^{2}=0.998). The sampling cost is 0.360.36 ns per element. The buffer cost is not resolvable from zero for k≤211k\leq 2^{11}, and equals 1.61.6 ns per element at k=214k=2^{14} and 2.42.4 ns at k=217k=2^{17}: selecting many elements from the buffer costs several times more per element than sampling, which is why the sample should grow with kk. The ratio is well described by 1/r⁡(k)=0.59​log2​k−4.31/r(k)=0.59\log_{2}k-4.3 for k>27k>2^{7} and r=∞r=\infty below, and this is the function r⁡(⋅)r(\cdot) used in Algorithm 1.

Figure 7: Exact-mode latency of Prof-K as a function of the sample size SS on the RTX 3060 for six (N,k)(N,k) configurations, with the sample size chosen by Algorithm 1 marked. The curves are flat over most of the range; at large kk and large NN a larger sample pays for itself through a smaller buffer, and the rule follows it.

Sensitivity and Accuracy of the Rule

Figure 7 shows six representative configurations and Table 5 summarizes all 4040. First, the latency is insensitive to SS over a wide range: across the 64×64\times span of the sweep, the ratio between the slowest and the fastest sample size has a median of 1.241.24 and never exceeds 1.511.51, so a factor of two in either direction around the optimum typically costs a few percent. Second, the empirical optimum is S=214S=2^{14} in 2727 of the 4040 configurations, which is the fixed-cost floor SminS_{\min}; the exceptions are the large-kk configurations, where the optimum moves to 2182^{18} as NN grows. Third, the sample size returned by Algorithm 1 lands on average 2.2%2.2\% above the optimal latency of the sweep, with a worst case of 15%15\% for moderate kk at large NN, where the rule leaves the floor slightly earlier than the measurements warrant; run-to-run variation between sessions on this GPU is of the same order, so we do not resolve the rule further.

Table 5: Sample-size ablation on the RTX 3060 (exact mode, ε=10−3\varepsilon=10^{-3}). For each NN, mean over the five values of kk of the ratio between the slowest and the fastest sample size in the sweep, and of the latency at the sample size chosen by Algorithm 1 relative to the empirical optimum.
NN 2212^{21} 2222^{22} 2232^{23} 2242^{24} 2252^{25} 2262^{26} 2272^{27} 2282^{28}
slowest / fastest SS 1.33 1.40 1.34 1.20 1.11 1.13 1.20 1.29
Algorithm 1 / optimum 1.01 1.00 1.01 1.02 1.04 1.04 1.05 1.01

E.2 Batched Inputs

The main evaluation concerns a single large vector. Here we ask whether Prof-K also pays off on batched inputs, where each of BB rows of length NN is selected independently in one call, and what relaxing the failure budget buys. We measure torch.topk, RadiK and Prof-K in approximate mode on the RTX 3060 for N∈{216,218,220,222}N\in\{2^{16},2^{18},2^{20},2^{22}\}, B∈{1,16,128,1024}B\in\{1,16,128,1024\} with B⋅N≤228B\cdot N\leq 2^{28}, and k∈{64,1024}k\in\{64,1024\}, on standard normal inputs of shape (B,N)(B,N), averaging over 55 input tensors and 2020 timed calls each after a warm-up. The budget ε\varepsilon is per call and is split as ε/B\varepsilon/B across the rows, so that by the union bound the probability that any row misses is at most ε\varepsilon; the recall, the fraction of the true top-kk indices returned, is measured against torch.topk.

Figure 8: Latency of torch.topk, RadiK and Prof-K (approximate mode, ε=10−3\varepsilon=10^{-3}) on batched inputs on the RTX 3060, for k=64k=64 (top) and k=1024k=1024 (bottom). Each group is one (N,B)(N,B) configuration, ordered by the total number of elements B⋅NB\cdot N; the number above a group is the speed-up of Prof-K over torch.topk.

Batching Extends the Gains to Small NN

Figure 8 shows the latencies at ε=10−3\varepsilon=10^{-3}. For a single row of N≤220N\leq 2^{20} no method takes more than 0.50.5 ms and the three are within 1010 to 40%40\% of each other, since the fixed costs of the four stages of Prof-K are comparable to an exact selection over a million elements. Batching amortizes these fixed costs, and the total size B⋅NB\cdot N is what governs the outcome. From B⋅N=223B\cdot N=2^{23} on, Prof-K is ahead of both baselines in every configuration, and its advantage grows with the total size: 3.4×3.4\times over torch.topk and 2.1×2.1\times over RadiK at N=220N=2^{20}, B=16B=16; 4.8×4.8\times and 2.6×2.6\times at N=222N=2^{22}, B=16B=16; 3.7×3.7\times and 2.0×2.0\times at N=220N=2^{20}, B=128B=128; and 2.5×2.5\times and 1.5×1.5\times at N=218N=2^{18}, B=1024B=1024, all for k=64k=64, with k=1024k=1024 between 1010 and 30%30\% lower. At a fixed total size, longer rows are more favourable than more rows, since the per-row costs of the threshold selection and of the buffer are then spread over more elements: B⋅N=226B\cdot N=2^{26} gives 4.8×4.8\times for N=222N=2^{22}, B=16B=16 but 1.8×1.8\times for N=216N=2^{16}, B=1024B=1024. The crossover at B⋅N≈222B\cdot N\approx 2^{22} coincides with the one for a single vector in Figure 2 once the horizontal axis there is read as the total size. The recall is 1.0001.000 in every configuration.

Relaxing the Budget

Table 6 reports the same measurements at four budgets. Loosening ε\varepsilon from 10−310^{-3} to 0.30.3 shortens the latency by only about 5%5\% at the median, because SS, t⋆t^{\star} and M⋆M^{\star} depend on ε\varepsilon only through z=Φ−1​(1−ε/2)z=\Phi^{-1}(1-\varepsilon/2), whereas the recall of single rows drops to between 0.870.87 and 0.970.97. On batches the split budget keeps the per-row budget small even at ε=0.3\varepsilon=0.3, and the recall stays above 0.9960.996. Prof-K therefore has no useful accuracy knob below the exact regime, and there is no reason to run it with a loose budget: the recommended operating point is ε=10−3\varepsilon=10^{-3}, at which the approximate mode returns the exact top-kk in all but one call in a thousand.

Table 6: Approximate mode of Prof-K on the RTX 3060 as a function of the failure budget ε\varepsilon, over the 2626 configurations of Figure 8: latency relative to ε=10−3\varepsilon=10^{-3} (median), and minimum and median recall for single rows (B=1B=1) and for batches (B≥16B\geq 16).
ε\varepsilon 10−310^{-3} 10−210^{-2} 0.10.1 0.30.3
latency relative to ε=10−3\varepsilon=10^{-3} 1.00 0.95 0.93 0.95
recall, B=1B=1 (min / median) 1.000 / 1.000 1.000 / 1.000 0.966 / 0.995 0.871 / 0.945
recall, B≥16B\geq 16 (min / median) 1.000 / 1.000 1.000 / 1.000 0.998 / 1.000 0.997 / 1.000

E.3 Adversarial Input Distributions

The guarantee of Section 4 depends only on the ranks induced by uniform sampling, never on the values, so the latency of Prof-K should be the same for every input distribution. Radix selection is not distribution-free: it discards candidates one digit of the binary representation at a time, and inputs whose values share their leading bits defeat the reduction. Li et al. [12] name values concentrated in a narrow range, such as uniform on [128.6,128.7][128.6,128.7], as an adversarial case for RadiK and counter it with an adaptive scaling step that subtracts a randomly drawn element before the selection. We measure all methods on the RTX 3060 at N=227N=2^{27} and k∈{28,213}k\in\{2^{8},2^{13}\} on seven input distributions: this adversarial case, uniform on [1,2)[1,2), whose values share their sign and exponent bits, and five benign ones, including the heavy-tailed log-normal and Cauchy. RadiK is run without its scaling step, as in the rest of the paper, and with it; Prof-K runs in exact mode at ε=10−3\varepsilon=10^{-3}, and every returned value is verified against torch.topk.

Table 7: Latency (ms) on the RTX 3060 at N=227N=2^{27} as a function of the input distribution: torch.topk, RadiK without and with (†\dagger) its adaptive scaling step, and Prof-K in exact mode at ε=10−3\varepsilon=10^{-3}. Mean over 55 input tensors and 2020 timed calls each.
k=28k=2^{8} k=213k=2^{13}
input distribution torch RadiK RadiK† Prof-K torch RadiK RadiK† Prof-K
standard normal 11.2 5.3 5.3 2.3 11.5 5.8 5.8 2.4
uniform [0,1)[0,1) 11.6 6.1 5.6 2.2 11.3 6.6 6.1 2.4
uniform [1,2)[1,2) 11.2 6.9 5.7 2.2 11.4 7.3 6.2 2.3
uniform [128.6,128.7][128.6,128.7] 10.8 15.8 5.6 2.2 11.2 16.4 6.0 2.4
log-normal 11.3 5.3 5.3 2.3 11.6 5.8 5.8 2.4
exponential 11.7 5.4 5.3 2.2 11.4 6.0 5.8 2.5
Cauchy 11.4 5.4 5.4 2.3 11.7 5.8 5.8 2.4

Table 7 confirms the picture. Prof-K takes 2.22.2 to 2.52.5 ms on every distribution, with at most 0.020.02 redraws per call and exact values on every call. RadiK without scaling slows down as soon as values share their leading bits, by 1515 to 30%30\% on uniform inputs and by 3×3\times on the narrow-range input, where it falls behind torch.topk; the speed-up of Prof-K over it grows accordingly, from 2.3×2.3\times on normal inputs to 7×7\times on the adversarial one. Its scaling step removes the worst case, after which Prof-K is still 2.32.3 to 2.6×2.6\times faster on every distribution. Heavy tails cost no method anything, since the top-kk elements then occupy their own leading digits. Prof-K is thus impervious to the input distribution in practice as well as in theory, and it needs no input-dependent preprocessing to be so.

E.4 Failure Budget

The failure budget ε\varepsilon is the one user-facing parameter of Prof-K. In exact mode it trades the cost of a single attempt, which grows slowly as ε\varepsilon shrinks through z=Φ−1​(1−ε/2)z=\Phi^{-1}(1-\varepsilon/2) in the sample size and the buffer capacity, against the frequency of redraws, which grows linearly in ε\varepsilon. We measure this trade-off on the RTX 3060 in exact mode with the exact parameters of Theorem 4, for ε\varepsilon from 0.30.3 down to 10−410^{-4} at N∈{224,228}N\in\{2^{24},2^{28}\} and k∈{28,214}k\in\{2^{8},2^{14}\}, recording for every call the count RR of the first attempt, the number of redraws and the latency of the whole call including the check. Between 2,0002{,}000 and 10,00010{,}000 calls are made per configuration, 140,000140{,}000 in total.

The Failure Probabilities Are Exact

The count RR of the first attempt tells which failure mode occurred, so the experiment measures Pr⁡(R<k)\Pr(R<k) and Pr⁡(R>M)\Pr(R>M) directly. Table 8 compares the observed numbers of recall failures and buffer overflows, pooled over the four (N,k)(N,k) configurations, with the numbers predicted by Propositions 2 and 3 at the same parameters. The two agree at every budget within the sampling noise of the counts, from 1,1941{,}194 overflows observed against 1,2001{,}200 predicted at ε=0.3\varepsilon=0.3 to 1515 against 2020 at 10−310^{-3}, and the observed total never exceeds ε\varepsilon times the number of calls. The recall failures are rarer than εA\varepsilon_{A} allows at small kk, where the integer threshold rank cannot exhaust its budget: at N=228N=2^{28}, k=28k=2^{8} and ε=0.3\varepsilon=0.3 the rank is t⋆=1t^{\star}=1 and Pr⁡(R<k)\Pr(R<k) is 0.0150.015 rather than 0.150.15.

Table 8: Predicted and observed numbers of recall failures (R<kR<k) and buffer overflows (R>MR>M) on the first attempt, pooled over the four (N,k)(N,k) configurations of the failure-budget ablation (RTX 3060, exact mode). Predictions are the exact probabilities of Propositions 2 and 3 multiplied by the number of calls.
ε\varepsilon calls predicted R<kR<k observed R<kR<k predicted R>MR>M observed R>MR>M
0.30.3 8,000 545.8 518 1,199.9 1,194
0.10.1 8,000 248.0 245 399.9 415
0.030.03 8,000 41.5 34 120.0 116
0.010.01 12,000 31.4 38 60.0 51
0.0030.003 24,000 16.5 25 36.0 30
10−310^{-3} 40,000 7.0 7 20.0 15
10−410^{-4} 40,000 1.0 0 2.0 2
Figure 9: Exact-mode latency of Prof-K on the RTX 3060 as a function of the failure budget ε\varepsilon, per call including the run-time check and any redraw. Left: mean latency relative to its minimum over ε\varepsilon. Right: the 9999th percentile relative to the median.

The Expected Cost Is Flat, but the Tail Events Are Not

Figure 9 shows the latency. The mean is within 10%10\% of its minimum for every ε\varepsilon between 10−310^{-3} and 10−210^{-2} in all four configurations, with the minimum between 0.0030.003 and 0.030.03: the larger buffer of a tight budget and the redraws of a loose one cost about the same there. Outside that range both ends lose, by up to 16%16\% at ε=10−4\varepsilon=10^{-4}, where the buffer at N=224N=2^{24}, k=28k=2^{8} is 34%34\% larger than at 10−310^{-3}, and by 1515 to 30%30\% at ε=0.3\varepsilon=0.3, where a call redraws 0.20.2 to 0.40.4 times on average. The tail separates the two ends: for ε≤10−2\varepsilon\leq 10^{-2} the 9999th percentile stays within the run-to-run jitter of this GPU, 1.31.3 to 2×2\times the median, whereas at ε=0.1\varepsilon=0.1 one call in ten runs the pass twice and the 9999th percentile rises to 2.22.2 to 3.6×3.6\times the median, and to 33 to 4.3×4.3\times at ε=0.3\varepsilon=0.3. A budget of 10−310^{-3} therefore costs at most 10%10\% in expectation against the fastest budget and keeps the latency of every call predictable, which is why we recommend it as a reasonable default.